/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/* Gateway kernel
 * Europagate, 1995
 *
 * $Log: gip.c,v $
 * Revision 1.8  1995/05/22 09:03:41  adam
 * New argument, block, to cs_open.
 *
 * Revision 1.7  1995/05/16  09:40:52  adam
 * LICENSE.
 *
 * Revision 1.6  1995/05/02  15:26:52  adam
 * EINTR obvserved on reads and writes.
 *
 * Revision 1.5  1995/05/01  16:27:28  adam
 * Various improvements. Close-on-exec and close on failure on either
 * read or write FIFO.
 *
 * Revision 1.4  1995/04/21  16:38:07  adam
 * Added more debugging logs.
 *
 * Revision 1.3  1995/04/20  15:12:42  adam
 * Minor hacks really.
 *
 * Revision 1.2  1995/03/27  12:52:18  adam
 * A little more verbose in marc dump.
 *
 * Revision 1.1  1995/03/27  08:24:58  adam
 * New module gip: Gateway IPc module.
 * New module gw-db: Gateway hash-db module (user information table).
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <errno.h>

#include <gw-log.h>

#include <gip.h>

GIP gip_initialize (const char *name)
{
    GIP gip = malloc (sizeof(*gip));

    if (!gip)
        return NULL;
    if (!(gip->name = malloc (strlen(name)+1)))
    {
        free (gip);
	return NULL;
    }
    strcpy (gip->name, name);
    gip->ret = mknod (gip->name, S_IFIFO|0666, 0);
    gip->errno = errno;
    gip->rfd = gip->wfd = -1;
    gw_log (GW_LOG_DEBUG, "gip", "Creating %s", gip->name);
    return gip;
}

int gip_destroy (GIP gip)
{
    assert (gip);

    unlink (gip->name);
    free (gip->name);
    free (gip);
    return 0;
}

int gip_infileno (GIP gip)
{
    return gip->rfd;
}

int gip_outfileno (GIP gip)
{
    return gip->wfd;
}

int gip_errno (GIP gip)
{
    return gip->errno;
}

int gip_read (GIP gip, char *buf, size_t count)
{
    int r, no = 0;
    while (no < count)
    {
        while (1)
        {
            r = read (gip->rfd, buf+no, count-no);
            if (r != -1)
                break;
            if (errno != EINTR)
            {
                gip->errno = errno;
                return -1;
            }
        }
	no += r;
    }
    return 0;
}

int gip_write (GIP gip, const char *buf, size_t count)
{
    int r, no = 0;
    while (no < count)
    {
        while (1)
        {
            r = write (gip->wfd, buf+no, count-no);
            if (r != -1)
                break;
            if (errno != EINTR)
            {
                gip->errno = errno;
                return -1;
            }
	}
	no += r;
    }
    return 0;
}

int gip_wline (GIP gip, const char *buf)
{
    return gip_write (gip, buf, strlen(buf));
}

