/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/* Gateway kernel
 * Europagate, 1995
 *
 * $Log: gips.c,v $
 * Revision 1.8  1995/05/22 11:43:01  adam
 * Minor changes on dtbsun - include of errno.h.
 *
 * Revision 1.7  1995/05/22  09:03:41  adam
 * New argument, block, to cs_open.
 *
 * Revision 1.6  1995/05/16  09:40:53  adam
 * LICENSE.
 *
 * Revision 1.5  1995/05/01  16:27:29  adam
 * Various improvements. Close-on-exec and close on failure on either
 * read or write FIFO.
 *
 * Revision 1.4  1995/04/21  16:38:08  adam
 * Added more debugging logs.
 *
 * Revision 1.3  1995/04/20  15:12:46  adam
 * Minor hacks really.
 *
 * Revision 1.2  1995/04/19  16:02:06  adam
 * Some hacks to get the FIFO communication work!! Isn't reliable.
 *
 * Revision 1.1  1995/03/27  08:25:00  adam
 * New module gip: Gateway IPc module.
 * New module gw-db: Gateway hash-db module (user information table).
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include <gw-log.h>
#include <gip.h>

static char *module = "gips";

GIP gips_initialize (const char *name)
{
    return gip_initialize (name);
}

int gips_destroy (GIP gip)
{
    return gip_destroy (gip);
}

int gips_open (GIP gip, const char *client, int block)
{
    do
    {
        gw_log (GW_LOG_DEBUG, module, "open readonly of %s", gip->name);
        gip->rfd = open (gip->name, block ? O_RDONLY : (O_RDONLY|O_NONBLOCK));
        gw_log (GW_LOG_DEBUG, module, "got rfd %d", gip->rfd);
    } while (gip->rfd == -1 && errno == EINTR);
    do
    {
        gw_log (GW_LOG_DEBUG, module, "open writeonly of %s", client);
        gip->wfd = open (client, block ? O_WRONLY : (O_WRONLY|O_NONBLOCK));
        gw_log (GW_LOG_DEBUG, module, "got wfd %d", gip->wfd);
    } while (gip->wfd == -1 && errno == EINTR);
    if (gip->rfd == -1)
    {
        gw_log (GW_LOG_DEBUG|GW_LOG_ERRNO, module, 
                "cannot open %s", gip->name);
        if (gip->wfd != -1)
        {
            close (gip->wfd);
            gip->wfd = -1;
        }
        else
            gw_log (GW_LOG_DEBUG|GW_LOG_ERRNO, module, 
                    "cannot open %s", client);
        return -1;
    }
    if (gip->wfd == -1)
    {
        gw_log (GW_LOG_DEBUG|GW_LOG_ERRNO, module, 
                "cannot open %s", client);
        close (gip->rfd);
        gip->rfd = -1;
        return -1;
    }
    fcntl (gip->wfd, F_SETFL, 0);
    fcntl (gip->rfd, F_SETFL, 0);
    fcntl (gip->wfd, F_SETFD, FD_CLOEXEC);
    fcntl (gip->rfd, F_SETFD, FD_CLOEXEC);
    return 0;
}

int gips_close (GIP gip)
{
    if (gip->rfd != -1)
    {
        close (gip->rfd);
	gip->rfd = -1;
    }
    if (gip->wfd != -1)
    {
        close (gip->wfd);
	gip->wfd = -1;
    }
    return 0;
}
