/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * $Log: wirtcl.c,v $
 * Revision 1.16  1996/02/29 15:40:23  adam
 * New function w_interp_irtcl_get that returns Tcl interpreter of
 * IrTcl interpreter.
 *
 * Revision 1.15  1996/02/21  14:58:01  adam
 * Modified to use ir_tcl_select_set.
 *
 * Revision 1.14  1996/02/12  10:10:31  adam
 * Resource/config system used by the gateway.
 *
 * Revision 1.13  1996/01/24  08:26:54  adam
 * All tcl commands prefixed with egw_ (except the html command).
 *
 * Revision 1.12  1996/01/12  10:05:18  adam
 * If script name ends with ';' HTTP/GET/Expires will be defined.
 * The cgi interface only reads final handshake if response from
 * server (shell) was zero-terminated [If it isn't it probably died].
 *
 * Revision 1.11  1996/01/09  16:16:49  adam
 * Port to OSF/1. Gif references moved from /gif/ to /egwgif/.
 *
 * Revision 1.10  1996/01/09  10:46:50  adam
 * New defines: LOGDIR/EGWDIR/CGIDIR set in Makefile.
 *
 * Revision 1.9  1995/11/07  14:56:59  adam
 * Work on search in multiple targets.
 * New wtcl command: wlog.
 * Optional timeout parameter to zwait.
 *
 * Revision 1.8  1995/11/06  17:44:22  adam
 * State reestablised when shell restarts. History of previous
 * result sets.
 *
 * Revision 1.7  1995/11/02  16:35:37  adam
 * Bug fixes and select on FIFOs in wcgi - doesn't really work!
 *
 * Revision 1.6  1995/11/01  16:15:47  adam
 * Better presentation of records. Query/set number persistent.
 *
 * Revision 1.5  1995/10/31  16:56:24  adam
 * Record presentation.
 *
 * Revision 1.4  1995/10/31  10:03:53  adam
 * Work on queries.
 * New command implemented - aborts script.
 *
 * Revision 1.3  1995/10/30  17:35:18  adam
 * New function zwait that waits for a variable change - due to i/o events
 * that invoke callback routines.
 *
 * Revision 1.2  1995/10/27  17:30:16  adam
 * First search request/response that works.
 *
 * Revision 1.1  1995/10/27  15:12:08  adam
 * IrTcl incorporated in the gateway.
 * Better separation of script types.
 * Z39.50 gateway scripts entered.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#ifdef AIX
#include <sys/select.h>
#endif
#include <string.h>
#include <assert.h>
#include <ctype.h>

#include <log.h>
#include "wtcl.h"
#include "wirtcl.h"

static void *do_create (WCLIENT wcl, void *args);
static int do_exec (const char *fname, char *parms, void *mydata);
static int do_load (char *parms, void *mydata);
static int do_save (char *parms, void *mydata);

static struct w_interp_type w_interp_t = {
    "irtcl",
    do_create,
    do_exec,
    do_load,
    do_save
};

W_Interp_Type w_interp_irtcl = &w_interp_t;


static char *mod = "wirtcl";

struct tcl_info {
    W_Interp w_interp;
    Tcl_Interp *interp;
    WCLIENT wcl;
};

static int events (struct tcl_info *p, char *waitVar, int tout);

static int proc_zwait_invoke (ClientData clientData, Tcl_Interp *interp,
                              int argc, char **argv)
{
    struct tcl_info *p = (struct tcl_info*) clientData;
    
    if (argc < 2)
        return TCL_OK;
    return events (p, argv[1], (argc == 3) ? atoi(argv[2]) : 0);
}

/* select callbacks */
struct callback {
    void (*handle)(ClientData, int, int, int);
    int r, w, e;
    ClientData obj;
};
#define MAX_CALLBACK 200

static struct callback callback_table[MAX_CALLBACK];
static int max_fd = 3;            /* don't worry: it will grow... */

static void *do_create (WCLIENT wcl, void *args)
{
    struct tcl_info *p;
    int i;

    if (!(p = malloc (sizeof(*p))))
    {
        gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, mod, "malloc: irtcl_info");
        exit (1);
    }
    if (!(p->w_interp = w_interp_create (w_interp_tcl, wcl, NULL)))
    {
        gw_log (GW_LOG_FATAL, mod, "Cannot make Tcl_Interp");
        exit (1);
    }
    p->wcl = wcl;
    p->interp = w_interp_tcl_get (p->w_interp);
    if (Irtcl_Init(p->interp) == TCL_ERROR)
    {
        gw_log (GW_LOG_FATAL, mod, "Cannot make Irtcl_Interp");
        exit (1);
    }
    log_init(LOG_ALL, "irtcl", "irtcl_log");
    /* initialize irtcl */
    Tcl_CreateCommand (p->interp, "egw_wait", proc_zwait_invoke, p, NULL);
    for (i=0; i<MAX_CALLBACK; i++)
        callback_table[i].handle = NULL;
    return p;
}


static int do_exec (const char *fname, char *parms, void *mydata)
{
    struct tcl_info *p = mydata;
    int r;
    if ((r = w_interp_exec (p->w_interp, fname, parms)))
        return r;
    return 0;
}


static int events (struct tcl_info *p, char *waitVar, int tout)
{
    int r, i;
    char *cp;
    char *waitVarVal;
    static fd_set fdset_tcl_r;
    static fd_set fdset_tcl_w;
    static fd_set fdset_tcl_x;
    int fifo_in = p->wcl->linein;
    if (fifo_in > max_fd)
        max_fd = fifo_in;

    assert (waitVar);
    if ((cp = Tcl_GetVar (p->interp, waitVar, 0)))
    {
        waitVarVal = malloc (strlen(cp)+1);
        strcpy (waitVarVal, cp);
    }
    else
    {
        char msg[128];

        sprintf (msg, "Variable %s doesn't exist", waitVar);
        gw_log (GW_LOG_WARN, mod, "%s", msg);
        Tcl_AppendResult (p->interp, msg, NULL);
        return TCL_ERROR;
    }
    gw_log (GW_LOG_DEBUG, mod, "Waiting %s=%s", waitVar, waitVarVal);
    while (1)
    {
        struct timeval to, *top;
	if (tout > 0)
	{
	    to.tv_usec = 0;
	    to.tv_sec = tout;
	    top = &to;
	}
	else
	    top = 0;

        if (!(cp = Tcl_GetVar (p->interp, waitVar, 0)) ||
            strcmp (cp, waitVarVal))
        {
            gw_log (GW_LOG_DEBUG, mod, "Changed to %s", cp);
            Tcl_AppendResult (p->interp, cp, NULL);
            free (waitVarVal);
            return TCL_OK;
        }
        FD_ZERO (&fdset_tcl_r);
        FD_ZERO (&fdset_tcl_w);
        FD_ZERO (&fdset_tcl_x);
        
        for (r=0, i=0; i<=max_fd; i++)
        {
            if (callback_table[i].handle && callback_table[i].w)
            {
                FD_SET (i, &fdset_tcl_w);
                r++;
            }
            if (callback_table[i].handle && callback_table[i].r)
            {
                FD_SET (i, &fdset_tcl_r);
                r++;
            }
            if (callback_table[i].handle && callback_table[i].e)
            {
                FD_SET (i, &fdset_tcl_x);
                r++;
            }
        }
        if (!r)
            break;
#if 1
        gw_log (GW_LOG_DEBUG, mod, "fifo select %d", fifo_in);
        FD_SET (fifo_in, &fdset_tcl_r);
#endif
        if ((r = select(max_fd+1, &fdset_tcl_r, &fdset_tcl_w, 
                          &fdset_tcl_x, top)) < 0)
        {
            gw_log (GW_LOG_ERRNO|GW_LOG_FATAL, mod, "select");
            exit(1);
        }
        if (!r)
        {
            gw_log (GW_LOG_DEBUG, mod, "timeout");
            free (waitVarVal);
            return TCL_ERROR;
        }
        if (FD_ISSET (fifo_in, &fdset_tcl_r))
        {
            gw_log (GW_LOG_DEBUG, mod, "FIFO closed");
            free (waitVarVal);
            return TCL_ERROR;
        }
        for (i=0; i<=max_fd; i++)
        {
            int r_flag = 0;
            int w_flag = 0;
            int e_flag = 0;

            if (!callback_table[i].handle)
                continue;
            if (FD_ISSET (i, &fdset_tcl_r) && callback_table[i].r)
                r_flag = 1;
            if (FD_ISSET (i, &fdset_tcl_w) && callback_table[i].w)
                w_flag = 1;
            if (FD_ISSET (i, &fdset_tcl_x) && callback_table[i].e)
                e_flag = 1;
            if (r_flag || w_flag || e_flag)
                (*callback_table[i].handle)(callback_table[i].obj,
                 r_flag, w_flag, e_flag);
        }
    }
    free (waitVarVal);
    return TCL_OK;
}

void ir_tcl_select_set (void (*f)(ClientData clientData, int r, int w, int e),
                        int fd, ClientData clientData, int r, int w, int e)
{
    callback_table[fd].handle = f;
    callback_table[fd].obj = clientData;
    callback_table[fd].r = r;
    callback_table[fd].w = w;
    callback_table[fd].e = e;
    if (fd > max_fd)
        max_fd = fd;
}

static int do_load (char *parms, void *mydata)
{
    struct tcl_info *p = mydata;

    return w_interp_load_state (p->w_interp, parms);
}

static int do_save (char *parms, void *mydata)
{
    struct tcl_info *p = mydata;

    return w_interp_save_state (p->w_interp, parms);
}

Tcl_Interp *w_interp_irtcl_get (W_Interp w_interp)
{
    struct tcl_info *p;

    if (strcmp (w_interp->ctrl->name, "irtcl"))
    {
        gw_log (GW_LOG_FATAL, mod, "Internal failure");
        assert (0);
    }
    p = (struct tcl_info*) w_interp->mydata;
    return p->interp;
}

