/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * $Log: wproto.c,v $
 * Revision 1.18  1996/02/12 10:09:23  adam
 * New parameter to wproto_init: directory root for the FIFOs (instead
 * of using preprocessor defines FIFODIR/FIFOROOT).
 *
 * Revision 1.17  1996/01/26  09:02:22  adam
 * Open of client FIFO called with O_NDELAY when reconnecting to shell
 * in order to prevent serious lock if previous shell died without
 * unlinking client FIFO.
 *
 * Revision 1.16  1996/01/24  10:13:56  adam
 * Bug fix: in function wo_write realloc is used only when memory is already
 * allocated with malloc.
 *
 * Revision 1.15  1996/01/24  08:25:32  adam
 * Buf fix: Uninitialized outbuffer_offset member.
 *
 * Revision 1.14  1996/01/12  13:08:07  adam
 * CGI script passes name of lock file to the shell. The server will not close
 * the response FIFO until this file becomes unlocked. This method handles
 * cancel operations much better.
 *
 * Revision 1.13  1996/01/12  10:05:20  adam
 * If script name ends with ';' HTTP/GET/Expires will be defined.
 * The cgi interface only reads final handshake if response from
 * server (shell) was zero-terminated [If it isn't it probably died].
 *
 * Revision 1.12  1996/01/05  16:35:02  adam
 * Minor changes.
 *
 * Revision 1.11  1996/01/05  16:21:21  adam
 * Bug fix: shell (wproto) sometimes closed server FIFO before cgi
 * program opened it - solution: cgi sends OK when response has been read.
 *
 * Revision 1.10  1995/12/22  14:21:16  adam
 * More work on scan. The search.egw script takes care of cached
 * query page (doesn't always increment nextSetNo). To make new search set
 * either 'New query' must be selected or the query page must be reloaded.
 * The msearch script doesn't do this yet, however.
 *
 * Revision 1.9  1995/11/14  16:31:36  adam
 * Temporary remove of ccl entry.
 *
 * Revision 1.8  1995/11/13  15:41:45  adam
 * Arrow gifs.
 * Gateway uses record element set names B(rief) and F(ull).
 * Bug fix. Didn't save idAuthentication correctly.
 *
 * Revision 1.7  1995/11/10  14:47:32  adam
 * Plus (+) characters automatically converted to space in forms.
 * Work on search in multiple targets. Doesn't work well - yet.
 * Presentation formats enhanced.
 *
 * Revision 1.6  1995/11/06  10:51:17  adam
 * End of response marker in response from wsh/wproto to wcgi.
 * Shells are respawned when necessary.
 *
 * Revision 1.5  1995/11/02  16:35:37  adam
 * Bug fixes and select on FIFOs in wcgi - doesn't really work!
 *
 * Revision 1.4  1995/10/31  16:56:25  adam
 * Record presentation.
 *
 * Revision 1.3  1995/10/27  15:12:10  adam
 * IrTcl incorporated in the gateway.
 * Better separation of script types.
 * Z39.50 gateway scripts entered.
 *
 * Revision 1.2  1995/10/23  16:55:39  adam
 * A lot of changes - really.
 *
 * Revision 1.1  1995/10/20  11:49:26  adam
 * First version of www gateway.
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdarg.h>
#include <ctype.h>
#include <errno.h>

#include "wproto.h"

static int wproto_dumpcache(WCLIENT wc, int level);
static int wproto_findcache(WCLIENT wc, char *name);
static void wproto_uncache(WCLIENT wc, int level);

static char *mod = "wproto";

void wo_write (WCLIENT wc, const char *s, size_t len)
{
    if (wc->outbuffer_offset + len >= wc->outbuffer_size)
    {
	if (!wc->outbuffer)
	    wc->outbuffer = malloc (wc->outbuffer_size += 
	       ((len >= OUTBUFFER_CHUNK) ? len*2 : OUTBUFFER_CHUNK));
	else
    	    wc->outbuffer = realloc(wc->outbuffer, wc->outbuffer_size +=
	       ((len >= OUTBUFFER_CHUNK) ? len*2 : OUTBUFFER_CHUNK));
    }
    memcpy(wc->outbuffer + wc->outbuffer_offset, s, len);
    wc->outbuffer_offset += len;
}

void wo_puts (WCLIENT wc, const char *s)
{
    wo_write (wc, s, strlen(s));
}

void wo_printf (WCLIENT wc, const char *fmt, ...)
{
    va_list ap;
    char tmpbuf[4048];

    va_start(ap, fmt);
    vsprintf(tmpbuf, fmt, ap);
    wo_puts(wc, tmpbuf);
    va_end(ap);
}

void wo_clear (WCLIENT wc, const char *type)
{
    wc->outbuffer_offset = 0;
    if (type)
        wo_printf(wc, "Content-type: %s\n\n", type);
}

int wo_puthtml (WCLIENT wc, char *name)
{
    FILE *f; 
    char ch;

    wo_clear(wc, "text/html");
    if (!(f = fopen(name, "r")))
    {
    	wo_printf(wc, "<BR>Failed to open file: %s<BR>", name);
    	return 0;
    }
    while (ch = getc(f), !feof(f))
    {
    	if (wo_putc(wc, ch) < 0)
    	{
	    fclose(f);
	    return -1;
	}
    }
    fclose(f);
    return 0;
}

int wo_flush(WCLIENT wc)
{
    int wrote, towrite;

    if (!(wc->outbuffer_offset))
	return 0;
    towrite = wc->outbuffer_offset;
    wc->outbuffer_offset = 0;
    for (;;)
    {
        int w_chunk;

        w_chunk = towrite;
	wrote = write(wc->lineout, wc->outbuffer + wc->outbuffer_offset,
	    w_chunk);
    	if (wrote <= 0)
    	{
	    gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, mod, "write response");
	    return -1;
	}
        gw_log (GW_LOG_DEBUG, mod, "wrote %d bytes", wrote);
	if (wc->cache_fd >= 0)
	    if (write(wc->cache_fd, wc->outbuffer + wc->outbuffer_offset,
		towrite) < 0)
	    {	
	    	gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, mod, "write cache");
	    	return -1;
	    }
	towrite -= wrote;
	if (!towrite)
	    break;
	wc->outbuffer_offset += wrote;
    }
    wc->outbuffer_offset = 0;
    return 0;
}

int wo_overflow(WCLIENT wc, char ch)
{
    gw_log (GW_LOG_DEBUG, mod, "wo_overflow");
    if (wo_flush(wc) < 0)
    	return -1;
    return wo_putc(wc, ch);
}

int wo_finish(WCLIENT wc)
{
    int fd;
    gw_log (GW_LOG_DEBUG, mod, "wo_finish");

    wo_putc (wc, 0);
    if (wo_flush(wc) < 0)
    	return -1;

    fd = open (wc->wf_serverf, O_RDONLY);
    if (fd != -1)
    {
        struct flock area;
        area.l_type = F_RDLCK;
        area.l_whence = SEEK_SET;
        area.l_start = 0L;
        area.l_len = 0L;
        fcntl (fd, F_SETLKW, &area);
        close (fd);
    }
    close(wc->lineout);
    wc->lineout = -1;
    if (wc->cache_fd >= 0)
    {
    	close(wc->cache_fd);
    	wc->cache_fd = -1;
    }
    return 0;
}

static void descramble(char *t, const char *o)
{
    unsigned int v;

    while (*o)
    {
    	if (*o == '%' && isxdigit(*(o + 1)) && isxdigit(*(o + 2)))
    	{
	    sscanf(o + 1, "%2x", &v);
	    o += 3;
            if (v == '+')
                *t = ' ';
            else
                *t = (char) v;
            t++;
	}
	else
        {
            if (*o == '+')
                *t = ' ';
            else
                *t = *o;
            t++;
            o++;
        }
    }
    *t = '\0';
}

static void decode_form(wform_data *form, char *buf)
{
    int i = 0;
    char *p;
    char tmp[512];

    while (*buf)
    {
    	for (p = form[i].name; *buf && *buf != '='; buf++)
	    *(p++) = *buf;
	*p = '\0';
	if (*buf)
	    buf++;
    	for (p = tmp; *buf && *buf != '&'; buf++)
	    *(p++) = *buf;
	*p = '\0';
	descramble(form[i].value, tmp);
	if (*buf)
	    buf++;
	i++;
    }
    *form[i].name = '\0';
}

char *wgetval(WCLIENT wc, char *name)
{
    int i;

    for (i = 0; *wc->wf_data[i].name; i++)
    	if (!strcmp(name, wc->wf_data[i].name))
	    return wc->wf_data[i].value;
    return 0;
}

int wproto_process(WCLIENT wc, int timeout)
{
    int toread, rs, level;
    char combuf[COMBUF], *p,*t;
    fd_set input;
    struct timeval to, *top;

    for (;;)
    {
	gw_log (GW_LOG_DEBUG, mod, "process waiting for input.");
	if (timeout > 0)
	{
	    to.tv_usec = 0;
	    to.tv_sec = timeout;
	    top = &to;
	}
	else
	    top = 0;
	FD_ZERO(&input);
	FD_SET(wc->linein, &input);
        /* go through select handle list */
	while ((rs = select(wc->linein + 1, &input, 0, 0, top)) < 0 &&
	    errno == EINTR)
	    ;
	if (rs < 0)
	{
	    gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, mod, "select");
	    return -1;
	}
	if (rs == 0)
	{
	    gw_log (GW_LOG_STAT, mod, 
                    "select %d second timeout.",
		    timeout);
            unlink (wc->wf_serverp);
	    return 0;
	}
	if (read(wc->linein, &toread, sizeof(toread)) < sizeof(toread))
	{
	    gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, mod, "wp_proc:len read failed");
	    exit(1);
	}
	toread -= sizeof(toread);
	if (read(wc->linein, combuf, toread) < toread)
	{
	    gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, mod, "wp_proc: data read failed");
	    exit(1);
	}
	p = combuf;
	for (t = wc->wf_serverp; (*t = *p); t++, p++);
	p++;
	for (t = wc->wf_serverf; (*t = *p); t++, p++);
	p++;
	for (t = wc->wf_parms; (*t = *p); t++, p++);
	p++;
	p++;         /* we don't deal with envvars yet */
	decode_form(wc->wf_data, p);
	if (wc->lineout < 0)
        {
            gw_log (GW_LOG_DEBUG, mod, "open %s", wc->wf_serverp);
            if ((wc->lineout = open(wc->wf_serverp, O_WRONLY)) < 0)
	    {
	        gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, mod, "open %s", 
                        wc->wf_serverp);
	        exit(1);
	    }
        }
	/* look in cache only if request carries no forms data. */
	if (!*wc->wf_data[0].name && (level = wproto_findcache(wc,
	    wc->wf_parms)) >= 0)
	{
	    gw_log (GW_LOG_DEBUG, mod, "wproto_dumpcache");
	    wproto_dumpcache(wc, level);
	    wo_finish(wc);
            
	}
	else
        {
	    return 1;
        }
    }
}

WCLIENT wproto_init (const char *fifoDir)
{
    char *val, path2[256];
    wclient_data *new;

    gw_log (GW_LOG_DEBUG, mod, "wproto_init");
    close(1);    /* release us from the wserver */
    if (!(new = malloc(sizeof(*new))))
    {
        gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, mod, "malloc");
        exit (1);
    }
    if (!(val = getenv ("GWID")))
    {
        gw_log (GW_LOG_FATAL, mod, "GWID not set");
        exit (1);
    }
    new->fifoDir = fifoDir;
    new->id = atoi (val);
    sprintf(new->path, "%s/clt%d", new->fifoDir, new->id);
    if (mkfifo(new->path, 0666 | S_IFIFO) < 0)
    	gw_log (GW_LOG_WARN|GW_LOG_ERRNO, mod, "mkfifo(%s)", new->path);
    gw_log (GW_LOG_DEBUG, mod, "Synchronizing with server.");
    sprintf(path2, "%s/srv%d", new->fifoDir, getppid());
    if ((new->lineout = open(path2, O_WRONLY)) < 0)
    {
    	gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, mod, "open out %s", path2);
    	exit(1);
    }
    if (write(new->lineout, "OK", 2) < 2)
    {
    	gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, mod, "write");
    	exit(1);
    }
    gw_log (GW_LOG_DEBUG, mod, "Synchronized.");
    if ((new->linein = open(new->path, O_RDONLY)) < 0)
    {
    	gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, mod, "open input %s", new->path);
    	exit(1);
    }
    gw_log (GW_LOG_DEBUG, mod, "init. linein=%d lineout=%d",
            new->linein, new->lineout);
    /* we put a handle on this so we get a blocking read when no peer */
    if (open(new->path, O_WRONLY | O_NDELAY) < 0)
    {
    	gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, mod, "open dummy %s", new->path);
    	exit(1);
    }
    new->outbuffer = 0;
    new->outbuffer_size = 0;
    new->outbuffer_offset = 0;
    new->cache_level = -1;
    new->cache_fd = -1;
    return new;
}

static void wproto_uncache(WCLIENT wc, int level)
{
    for (;wc->cache_level >= level; wc->cache_level--)
    	unlink(wc->cache[wc->cache_level].path);
}

void wproto_terminate(WCLIENT wc)
{
    close(wc->linein);
    unlink(wc->path);
    wproto_uncache(wc, 0);
    free(wc);
}

int wproto_cache(WCLIENT wc, int level)
{
    cache_data *p;

    if (level > wc->cache_level + 1)
    {
    	gw_log (GW_LOG_FATAL, mod, "Illegal cache level increment.");
    	exit(1);
    }
    wproto_uncache(wc, level);
    p = &wc->cache[++wc->cache_level];
    sprintf(p->path, "%s/csh%d.%d", wc->fifoDir, wc->id, level);
    if ((wc->cache_fd = open(p->path, O_WRONLY|O_CREAT|O_TRUNC, 0600)) < 0)
    {
    	gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, mod, "open %s", p->path);
    	return -1;
    }
    strcpy(p->name, wc->wf_parms);
    return 0;
}

static int wproto_findcache(WCLIENT wc, char *name)
{
    int i;

    for (i = 0; i <= wc->cache_level; i++)
    	if (!strcmp(wc->cache[i].name, name))
	    return i;
    return -1;
}

static int wproto_dumpcache(WCLIENT wc, int level)
{
    int fd, rd;

    gw_log (GW_LOG_STAT, mod, "Using Cache: %s", wc->cache[level].name);
    if ((fd = open(wc->cache[level].path, O_RDONLY)) < 0)
    {
    	gw_log (GW_LOG_FATAL, mod, "open (R) %s", wc->cache[level].path);
    	return -1;
    }
    while ((rd = read(fd, wc->outbuffer, OUTBUFFER_CHUNK)) > 0)
    	if (write(wc->lineout, wc->outbuffer, rd) < rd)
    	{
	    gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, mod, "write toline");
	    return -1;
	}
    if (rd < 0)
    {
    	gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, mod, "read");
    	return -1;
    }
    wproto_uncache(wc, level + 1);
    return 0;
}
